/**
 * @file cx16.h
 * @author Jesper Gravgaard / Sven Van de Velde
 * @brief Contains functions to control the features of the Commander X16.
 * Commander X16 Functions.
 * https://www.commanderx16.com/forum/index.php?/about-faq/
 * https://github.com/commanderx16/x16-docs/blob/master/Commander%20X16%20Programmer's%20Reference%20Guide.md
 * @version 0.2
 * @date 2022-01-29
 *
 * @copyright Copyright (c) 2022
 *
 */
.cpu _65c02
  // Commander X16 PRG executable file
.file [name="cx16-input.prg", type="prg", segments="Program"]
.segmentdef Program [segments="Basic, Code, Data"]
.segmentdef Basic [start=$0801]
.segmentdef Code [start=$80d]
.segmentdef Data [startAfter="Code"]
.segment Basic
:BasicUpstart(__start)
  .const WHITE = 1
  .const BLUE = 6
  ///< Read a character from the current channel for input.
  .const CBM_GETIN = $ffe4
  ///< Close a logical file.
  .const CBM_CLRCHN = $ffcc
  ///< Load a logical file.
  .const CBM_PLOT = $fff0
  .const VERA_INC_1 = $10
  .const VERA_ADDRSEL = 1
  .const VERA_LAYER_WIDTH_MASK = $30
  .const VERA_LAYER_HEIGHT_MASK = $c0
  .const OFFSET_STRUCT_PRINTF_BUFFER_NUMBER_DIGITS = 1
  .const STACK_BASE = $103
  .const isr_vsync = $314
  .const SIZEOF_STRUCT___0 = $8f
  .const SIZEOF_STRUCT_PRINTF_BUFFER_NUMBER = $c
  /// $9F20 VRAM Address (7:0)
  .label VERA_ADDRX_L = $9f20
  /// $9F21 VRAM Address (15:8)
  .label VERA_ADDRX_M = $9f21
  /// $9F22 VRAM Address (7:0)
  /// Bit 4-7: Address Increment  The following is the amount incremented per value value:increment
  ///                             0:0, 1:1, 2:2, 3:4, 4:8, 5:16, 6:32, 7:64, 8:128, 9:256, 10:512, 11:40, 12:80, 13:160, 14:320, 15:640
  /// Bit 3: DECR Setting the DECR bit, will decrement instead of increment by the value set by the 'Address Increment' field.
  /// Bit 0: VRAM Address (16)
  .label VERA_ADDRX_H = $9f22
  /// $9F23	DATA0	VRAM Data port 0
  .label VERA_DATA0 = $9f23
  /// $9F24	DATA1	VRAM Data port 1
  .label VERA_DATA1 = $9f24
  /// $9F25	CTRL Control
  /// Bit 7: Reset
  /// Bit 1: DCSEL
  /// Bit 2: ADDRSEL
  .label VERA_CTRL = $9f25
  /// $9F2A	DC_HSCALE (DCSEL=0)	Active Display H-Scale
  .label VERA_DC_HSCALE = $9f2a
  /// $9F2B	DC_VSCALE (DCSEL=0)	Active Display V-Scale
  .label VERA_DC_VSCALE = $9f2b
  /// $9F34	L1_CONFIG   Layer 1 Configuration
  .label VERA_L1_CONFIG = $9f34
  /// $9F35	L1_MAPBASE	    Layer 1 Map Base Address (16:9)
  .label VERA_L1_MAPBASE = $9f35
  .label BRAM = 0
  .label BROM = 1
.segment Code
__start: {
    lda #0
    sta.z BRAM
    lda #4
    sta.z BROM
    jsr conio_x16_init
    jsr main
    rts
}
/// Set initial screen values.
conio_x16_init: {
    .label __4 = $12
    .label __6 = $12
    jsr screenlayer1
    jsr textcolor
    jsr bgcolor
    jsr cursor
    jsr cbm_k_plot_get
    lda.z __4+1
    sta __conio
    jsr cbm_k_plot_get
    lda.z __6
    sta __conio+1
    ldx __conio
    tay
    jsr gotoxy
    lda #1
    sta __conio+$f
    sta __conio+$f+1
    rts
}
// Output one character at the current cursor position
// Moves the cursor forward. Scrolls the entire screen if needed
// void cputc(__register(X) char c)
cputc: {
    .const OFFSET_STACK_C = 0
    tsx
    lda STACK_BASE+OFFSET_STACK_C,x
    tax
    cpx #'\n'
    beq __b1
    lda #VERA_ADDRSEL^$ff
    and VERA_CTRL
    sta VERA_CTRL
    lda __conio+$13
    sta VERA_ADDRX_L
    lda __conio+$13+1
    sta VERA_ADDRX_M
    lda #VERA_INC_1
    ora __conio+5
    sta VERA_ADDRX_H
    stx VERA_DATA0
    lda __conio+$d
    sta VERA_DATA0
    ldy __conio+2
    lda __conio+$11,y
    cmp #0
    beq __b5
    lda __conio
    cmp __conio+8
    bcs __b6
    inc __conio
  __b7:
    inc __conio+$13
    bne !+
    inc __conio+$13+1
  !:
    inc __conio+$13
    bne !+
    inc __conio+$13+1
  !:
    rts
  __b6:
    jsr cputln
    jmp __b7
  __b5:
    lda __conio
    cmp __conio+6
    bcs __b8
    inc __conio
    inc __conio+$13
    bne !+
    inc __conio+$13+1
  !:
    inc __conio+$13
    bne !+
    inc __conio+$13+1
  !:
    rts
  __b8:
    jsr cputln
    rts
  __b1:
    jsr cputln
    rts
}
// Find the value of a keypress as returned by conio kbhit()
main: {
    .label kbhit1_return = $17
    .label test = $17
    lda #<cputc
    sta.z printf_str.putc
    lda #>cputc
    sta.z printf_str.putc+1
    lda #<s
    sta.z printf_str.s
    lda #>s
    sta.z printf_str.s+1
    jsr printf_str
    lda #0
    sta.z test
  __b1:
    lda.z test
    beq kbhit1
    lda #<cputc
    sta.z printf_str.putc
    lda #>cputc
    sta.z printf_str.putc+1
    lda #<s1
    sta.z printf_str.s
    lda #>s1
    sta.z printf_str.s+1
    jsr printf_str
    ldx.z test
    jsr printf_uchar
    lda #<cputc
    sta.z printf_str.putc
    lda #>cputc
    sta.z printf_str.putc+1
    lda #<s2
    sta.z printf_str.s
    lda #>s2
    sta.z printf_str.s+1
    jsr printf_str
    rts
  kbhit1:
    jsr CBM_CLRCHN
    jsr cbm_k_getin
    sta.z kbhit1_return
    jmp __b1
  .segment Data
    s: .text @"\npress a key\n"
    .byte 0
    s1: .text @"\nchar = "
    .byte 0
    s2: .text @"\n"
    .byte 0
}
.segment Code
// Set the layer with which the conio will interact.
screenlayer1: {
    ldx VERA_L1_MAPBASE
    lda VERA_L1_CONFIG
    sta.z screenlayer.config
    jsr screenlayer
    rts
}
// Set the front color for text output. The old front text color setting is returned.
// - color: a 4 bit value ( decimal between 0 and 15).
//   This will only work when the VERA is in 16 color mode!
//   Note that on the VERA, the transparent color has value 0.
// char textcolor(char color)
textcolor: {
    lda #$f0
    and __conio+$d
    ora #WHITE
    sta __conio+$d
    rts
}
// Set the back color for text output.
// - color: a 4 bit value ( decimal between 0 and 15).
//   This will only work when the VERA is in 16 color mode!
//   Note that on the VERA, the transparent color has value 0.
// char bgcolor(char color)
bgcolor: {
    lda #$f
    and __conio+$d
    ora #BLUE<<4
    sta __conio+$d
    rts
}
// If onoff is 1, a cursor is displayed when waiting for keyboard input.
// If onoff is 0, the cursor is hidden when waiting for keyboard input.
// The function returns the old cursor setting.
// char cursor(char onoff)
cursor: {
    .const onoff = 0
    lda #onoff
    sta __conio+$c
    rts
}
/**
 * @brief Get current x and y cursor position.
 * @return An unsigned int where the hi byte is the x coordinate and the low byte is the y coordinate of the screen position.
 */
cbm_k_plot_get: {
    .label return = $12
    lda #0
    sta x
    sta y
    sec
        jsr CBM_PLOT
        stx y
        sty x
    
    lda x
    sta.z return+1
    lda y
    sta.z return
    rts
  .segment Data
    x: .byte 0
    y: .byte 0
}
.segment Code
// Set the cursor to the specified position
// void gotoxy(__register(X) char x, __register(Y) char y)
gotoxy: {
    .label __9 = 9
    cpx __conio+6
    bcs __b1
    jmp __b2
  __b1:
    ldx __conio+6
  __b2:
    stx __conio
    cpy __conio+7
    bcs __b3
    tya
  __b5:
    sta __conio+1
    lda __conio
    asl
    tax
    tya
    asl
    tay
    txa
    clc
    adc __conio+$15,y
    sta.z __9
    lda __conio+$15+1,y
    adc #0
    sta.z __9+1
    lda.z __9
    sta __conio+$13
    lda.z __9+1
    sta __conio+$13+1
    rts
  __b3:
    lda __conio+7
    jmp __b5
}
// Print a newline
cputln: {
    lda #0
    sta __conio
    inc __conio+1
    lda __conio+1
    asl
    tay
    lda __conio+$15,y
    sta __conio+$13
    lda __conio+$15+1,y
    sta __conio+$13+1
    ldy __conio+2
    lda __conio+$f,y
    cmp #0
    beq __breturn
    jsr cscroll
  __breturn:
    rts
}
/// Print a NUL-terminated string
// void printf_str(__zp($f) void (*putc)(char), __zp($d) const char *s)
printf_str: {
    .label s = $d
    .label putc = $f
  __b1:
    ldy #0
    lda (s),y
    inc.z s
    bne !+
    inc.z s+1
  !:
    cmp #0
    bne __b2
    rts
  __b2:
    pha
    jsr icall1
    pla
    jmp __b1
  icall1:
    jmp (putc)
}
// Print an unsigned char using a specific format
// void printf_uchar(void (*putc)(char), __register(X) char uvalue, char format_min_length, char format_justify_left, char format_sign_always, char format_zero_padding, char format_upper_case, char format_radix)
printf_uchar: {
    .label putc = cputc
    // Handle any sign
    lda #0
    sta printf_buffer
  // Format number into buffer
    jsr uctoa
    lda printf_buffer
  // Print using format
    jsr printf_number_buffer
    rts
}
/**
 * @brief Scan a character from keyboard without pressing enter.
 * 
 * @return char The character read.
 */
cbm_k_getin: {
    lda #0
    sta ch
    jsr CBM_GETIN
    sta ch
    rts
  .segment Data
    ch: .byte 0
}
.segment Code
// --- layer management in VERA ---
// void screenlayer(char layer, __register(X) char mapbase, __zp($18) char config)
screenlayer: {
    .label __2 = $15
    .label mapbase_offset = $12
    .label config = $18
    lda VERA_DC_HSCALE
    sta vera_dc_hscale_temp
    lda VERA_DC_VSCALE
    sta vera_dc_vscale_temp
    lda #0
    sta __conio+2
    txa
    rol
    rol
    and #1
    sta __conio+5
    txa
    asl
    ldy #0
    sta.z __2+1
    sty.z __2
    tya
    sta __conio+3
    lda.z __2+1
    sta __conio+3+1
    lda #VERA_LAYER_WIDTH_MASK
    and.z config
    lsr
    lsr
    lsr
    lsr
    tax
    lda VERA_LAYER_DIM,x
    sta __conio+8
    lda #VERA_LAYER_HEIGHT_MASK
    and.z config
    rol
    rol
    rol
    and #3
    tay
    lda VERA_LAYER_DIM,y
    sta __conio+9
    txa
    asl
    // __conio.rowshift = ((config & VERA_LAYER_WIDTH_MASK)>>4)+6;
    tay
    lda VERA_LAYER_SKIP,y
    sta __conio+$a
    lda VERA_LAYER_SKIP+1,y
    sta __conio+$a+1
    lda vera_dc_hscale_temp
    eor #$80
    beq !+
    lda #1
  !:
    eor #1
    tax
    lda #$28
    cpx #0
    beq !e+
  !:
    asl
    dex
    bne !-
  !e:
    sec
    sbc #1
    sta __conio+6
    lda vera_dc_vscale_temp
    eor #$80
    beq !+
    lda #1
  !:
    eor #1
    tax
    lda #$1e
    cpx #0
    beq !e+
  !:
    asl
    dex
    bne !-
  !e:
    sec
    sbc #1
    sta __conio+7
    lda __conio+3
    sta.z mapbase_offset
    lda __conio+3+1
    sta.z mapbase_offset+1
    ldx #0
  __b1:
    lda __conio+7
    stx.z $ff
    cmp.z $ff
    bcs __b2
    rts
  __b2:
    txa
    asl
    tay
    lda.z mapbase_offset
    sta __conio+$15,y
    lda.z mapbase_offset+1
    sta __conio+$15+1,y
    clc
    lda.z mapbase_offset
    adc __conio+$a
    sta.z mapbase_offset
    lda.z mapbase_offset+1
    adc __conio+$a+1
    sta.z mapbase_offset+1
    inx
    jmp __b1
  .segment Data
    VERA_LAYER_DIM: .byte $1f, $3f, $7f, $ff
    VERA_LAYER_SKIP: .word $40, $80, $100, $200
    vera_dc_hscale_temp: .byte 0
    vera_dc_vscale_temp: .byte 0
}
.segment Code
// Scroll the entire screen if the cursor is beyond the last line
cscroll: {
    lda __conio+7
    cmp __conio+1
    bcs __breturn
    ldy __conio+2
    lda __conio+$f,y
    cmp #0
    bne __b4
    lda __conio+7
    cmp __conio+1
    bcs __breturn
    ldy #0
    ldx #0
    jsr gotoxy
  __breturn:
    rts
  __b4:
    jsr insertup
    ldy __conio+7
    ldx #0
    jsr gotoxy
    jsr clearline
    rts
}
// Converts unsigned number value to a string representing it in RADIX format.
// If the leading digits are zero they are not included in the string.
// - value : The number to be converted to RADIX
// - buffer : receives the string representing the number and zero-termination.
// - radix : The radix to convert the number to (from the enum RADIX)
// void uctoa(__register(X) char value, __zp($f) char *buffer, char radix)
uctoa: {
    .const max_digits = 3
    .label digit_value = $c
    .label buffer = $f
    .label digit = $11
    .label started = $14
    lda #<printf_buffer+OFFSET_STRUCT_PRINTF_BUFFER_NUMBER_DIGITS
    sta.z buffer
    lda #>printf_buffer+OFFSET_STRUCT_PRINTF_BUFFER_NUMBER_DIGITS
    sta.z buffer+1
    lda #0
    sta.z started
    sta.z digit
  __b1:
    lda.z digit
    cmp #max_digits-1
    bcc __b2
    lda DIGITS,x
    ldy #0
    sta (buffer),y
    inc.z buffer
    bne !+
    inc.z buffer+1
  !:
    lda #0
    tay
    sta (buffer),y
    rts
  __b2:
    ldy.z digit
    lda RADIX_DECIMAL_VALUES_CHAR,y
    sta.z digit_value
    lda.z started
    bne __b5
    cpx.z digit_value
    bcs __b5
  __b4:
    inc.z digit
    jmp __b1
  __b5:
    lda.z buffer
    sta.z uctoa_append.buffer
    lda.z buffer+1
    sta.z uctoa_append.buffer+1
    jsr uctoa_append
    inc.z buffer
    bne !+
    inc.z buffer+1
  !:
    lda #1
    sta.z started
    jmp __b4
}
// Print the contents of the number buffer using a specific format.
// This handles minimum length, zero-filling, and left/right justification from the format
// void printf_number_buffer(void (*putc)(char), __register(A) char buffer_sign, char *buffer_digits, char format_min_length, char format_justify_left, char format_sign_always, char format_zero_padding, char format_upper_case, char format_radix)
printf_number_buffer: {
    .label buffer_digits = printf_buffer+OFFSET_STRUCT_PRINTF_BUFFER_NUMBER_DIGITS
    cmp #0
    beq __b2
    pha
    jsr cputc
    pla
  __b2:
    lda #<printf_uchar.putc
    sta.z printf_str.putc
    lda #>printf_uchar.putc
    sta.z printf_str.putc+1
    lda #<buffer_digits
    sta.z printf_str.s
    lda #>buffer_digits
    sta.z printf_str.s+1
    jsr printf_str
    rts
}
// Insert a new line, and scroll the upper part of the screen up.
// void insertup(char rows)
insertup: {
    .label width = $b
    .label y = 5
    lda __conio+6
    inc
    // {asm{.byte $db}}
    asl
    sta.z width
    lda #0
    sta.z y
  __b1:
    lda.z y
    cmp __conio+1
    bcc __b2
    jsr clearline
    rts
  __b2:
    ldx.z y
    inx
    lda.z y
    asl
    tay
    txa
    asl
    tax
    lda __conio+5
    sta.z memcpy8_vram_vram.dbank_vram
    lda __conio+$15,y
    sta.z memcpy8_vram_vram.doffset_vram
    lda __conio+$15+1,y
    sta.z memcpy8_vram_vram.doffset_vram+1
    ldy __conio+5
    lda __conio+$15,x
    sta.z memcpy8_vram_vram.soffset_vram
    lda __conio+$15+1,x
    sta.z memcpy8_vram_vram.soffset_vram+1
    lda.z width
    sta.z memcpy8_vram_vram.num8
    jsr memcpy8_vram_vram
    inc.z y
    jmp __b1
}
clearline: {
    .label addr = 6
    lda __conio+1
    asl
    tay
    lda __conio+$15,y
    sta.z addr
    lda __conio+$15+1,y
    sta.z addr+1
    lda #VERA_ADDRSEL^$ff
    and VERA_CTRL
    sta VERA_CTRL
    lda.z addr
    sta VERA_ADDRX_L
    lda.z addr+1
    sta VERA_ADDRX_M
    lda #VERA_INC_1
    ora __conio+5
    sta VERA_ADDRX_H
    ldx __conio+6
  __b1:
    lda #' '
    sta VERA_DATA0
    lda __conio+$d
    sta VERA_DATA0
    dex
    cpx #0
    bne __b1
    rts
}
// Used to convert a single digit of an unsigned number value to a string representation
// Counts a single digit up from '0' as long as the value is larger than sub.
// Each time the digit is increased sub is subtracted from value.
// - buffer : pointer to the char that receives the digit
// - value : The value where the digit will be derived from
// - sub : the value of a '1' in the digit. Subtracted continually while the digit is increased.
//        (For decimal the subs used are 10000, 1000, 100, 10, 1)
// returns : the value reduced by sub * digit so that it is less than sub.
// __register(X) char uctoa_append(__zp($d) char *buffer, __register(X) char value, __zp($c) char sub)
uctoa_append: {
    .label buffer = $d
    .label sub = $c
    ldy #0
  __b1:
    cpx.z sub
    bcs __b2
    lda DIGITS,y
    ldy #0
    sta (buffer),y
    rts
  __b2:
    iny
    txa
    sec
    sbc.z sub
    tax
    jmp __b1
}
/**
 * @brief Copy a block of memory in VRAM from a source to a target destination.
 * This function is designed to copy maximum 255 bytes of memory in one step.
 * If more than 255 bytes need to be copied, use the memcpy_vram_vram function.
 *
 * @see memcpy_vram_vram
 *
 * @param dbank_vram Bank of the destination location in vram.
 * @param doffset_vram Offset of the destination location in vram.
 * @param sbank_vram Bank of the source location in vram.
 * @param soffset_vram Offset of the source location in vram.
 * @param num16 Specified the amount of bytes to be copied.
 */
// void memcpy8_vram_vram(__zp(8) char dbank_vram, __zp(6) unsigned int doffset_vram, __register(Y) char sbank_vram, __zp(3) unsigned int soffset_vram, __register(X) char num8)
memcpy8_vram_vram: {
    .label dbank_vram = 8
    .label doffset_vram = 6
    .label soffset_vram = 3
    .label num8 = 2
    lda #VERA_ADDRSEL^$ff
    and VERA_CTRL
    sta VERA_CTRL
    lda.z soffset_vram
    sta VERA_ADDRX_L
    lda.z soffset_vram+1
    sta VERA_ADDRX_M
    tya
    ora #VERA_INC_1
    sta VERA_ADDRX_H
    lda #VERA_ADDRSEL
    ora VERA_CTRL
    sta VERA_CTRL
    lda.z doffset_vram
    sta VERA_ADDRX_L
    lda.z doffset_vram+1
    sta VERA_ADDRX_M
    lda #VERA_INC_1
    ora.z dbank_vram
    sta VERA_ADDRX_H
  __b1:
  // the size is only a byte, this is the fastest loop!
    ldx.z num8
    dex
    lda.z num8
    bne __b2
    rts
  __b2:
    lda VERA_DATA0
    sta VERA_DATA1
    stx.z num8
    jmp __b1
}
.segment Data
  // The digits used for numbers
  DIGITS: .text "0123456789abcdef"
  // Values of decimal digits
  RADIX_DECIMAL_VALUES_CHAR: .byte $64, $a
  __conio: .fill SIZEOF_STRUCT___0, 0
  // Buffer used for stringified number being printed
  printf_buffer: .fill SIZEOF_STRUCT_PRINTF_BUFFER_NUMBER, 0
